.button,
button {
    background-color: transparent;
    border: 1px solid rgb(209, 213, 219);
    box-sizing: border-box;
    color: white;
    color: var(--font-color);
    font-size: .875rem;
    line-height: 1.25rem;
    text-align: center;
    box-shadow: 0 2px 2px 0 rgba(0, 0, 0, 0.1);
    cursor: pointer;
    user-select: none;
    -webkit-user-select: none;
    touch-action: manipulation;
    border-radius: 5px;
    border-radius: var(--small-radius);
}

button {
    padding: 0.2rem 0.5rem;
}

.button {
    padding: 0.45rem 1rem;
}

.button:hover:enabled,
button:hover:enabled {
    background-color: rgb(249, 250, 251);
    color: #1F2833;
}

.button:active:enabled,
button:active:enabled {
    background-color: rgb(192, 192, 192);
    color: #1F2833;
}

.button:disabled,
button:disabled {
    opacity: 0.5;
    cursor: auto;
}

.button:focus,
button:focus {
    outline: 2px solid transparent;
    outline-offset: 2px;
}

.button:focus-visible,
button:focus-visible {
    box-shadow: none;
}

button.delete,
.button.delete {
    font-size: large;
    font-weight: 900;
    background-color: red;
    color: white;
}

button.delete:hover,
.button.delete:hover {
    background-color: white;
    color: red;
}

button.delete:active,
.button.delete:active {
    background-color: red;
    color: red;
}

.light-dark-toggle {
    background-color: transparent;
    color: var(--font-color);
}

.flex-row {
    display: flex;
    flex-direction: row;
    flex-wrap: wrap;
    align-items: end;
    gap: 10px;
    gap: var(--panel-padding);
}

.flex-row .dash-dropdown,
.flex-row .dash-input {
    min-width: 12rem;
    flex: 1;
}

.stretch {
    flex: 1;
}